from apidaora import appdaora, route


@route.get('/app-options')
async def app_options() -> str:
    return 'Options'


@route.get('/route-options', options=True)
async def route_options() -> str:
    return 'Options'


app = appdaora([app_options, route_options], options=True)



import functools
import itertools
from asyncio import iscoroutine
from dataclasses import is_dataclass
from http import HTTPStatus
from json import JSONDecodeError
from typing import (  # type: ignore
    Any,
    Awaitable,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    _GenericAlias,
)

import orjson
from jsondaora import as_typed_dict, dataclass_asjson, typed_dict_asjson
from jsondaora.deserializers import deserialize_field
from jsondaora.exceptions import DeserializationError

from ..asgi.base import (
    ASGIBody,
    ASGICallableResults,
    ASGIHeaders,
    ASGIPathArgs,
    ASGIQueryDict,
    ASGIResponse,
)
from ..asgi.responses import (
    make_html_response,
    make_json_response,
    make_no_content_response,
    make_not_found_response,
    make_text_response,
)
from ..asgi.router import Controller, Route
from ..bodies import GZipFactory
from ..content import ContentType
from ..exceptions import BadRequestError, InvalidReturnError
from ..header import Header
from ..method import MethodType
from ..middlewares import (
    MiddlewareRequest,
    Middlewares,
    make_asgi_input_from_requst,
)
from ..responses import Response
from .controller_input import controller_input


RESPONSES_MAP: Dict[
    Union[ContentType, HTTPStatus], Union[Callable[..., ASGIResponse]],
] = {
    ContentType.APPLICATION_JSON: make_json_response,
    ContentType.TEXT_PLAIN: make_text_response,
    ContentType.TEXT_HTML: make_html_response,
    HTTPStatus.NOT_FOUND: make_not_found_response,
    HTTPStatus.NO_CONTENT: make_no_content_response,
}


def make_route(
    path_pattern: str,
    method: MethodType,
    controller: Union[Callable[..., Any], Controller],
    has_content_length: bool = True,
    route_middlewares: Optional[Union[Middlewares]] = None,
    options: bool = False,
) -> Route:
    ControllerInput = controller_input(controller, path_pattern)
    annotations_info = ControllerInput.__annotations_info__
    annotations_path_args = ControllerInput.__annotations_path_args__
    annotations_query_dict = ControllerInput.__annotations_query_dict__
    annotations_headers = ControllerInput.__annotations_headers__
    body_type = ControllerInput.__annotations_body__.get('body')
    return_type = controller.__annotations__.get('return')

    def parse_asgi_input(
        path_args: ASGIPathArgs,
        query_dict: ASGIQueryDict,
        headers: ASGIHeaders,
        body: ASGIBody,
        middlewares: Optional[Middlewares] = None,
    ) -> Dict[str, Any]:
        kwargs: Dict[str, Any]
        middleware_request: MiddlewareRequest

        if middlewares:
            middleware_request = MiddlewareRequest(path_pattern)

        if annotations_info.has_input:
            if annotations_info.has_path_args:
                kwargs = {
                    name: path_args.get(name)
                    for name in annotations_path_args.keys()
                }

                if middlewares:
                    middleware_request.path_args = {
                        name: deserialize_field(
                            name, type_, path_args.get(name)
                        )
                        for name, type_ in annotations_path_args.items()
                    }

            else:
                kwargs = {}

            if annotations_info.has_query_dict:
                kwargs.update(
                    {
                        name: parse_query_args(
                            name, type_, query_dict.get(name)
                        )
                        for name, type_ in annotations_query_dict.items()
                    }
                )

                if middlewares:
                    middleware_request.query_dict = {
                        name: deserialize_field(
                            name,
                            type_,
                            parse_query_args(
                                name, type_, query_dict.get(name)
                            ),
                        )
                        for name, type_ in annotations_query_dict.items()
                        if name != 'body'
                    }

            if annotations_info.has_headers:
                headers_map = ControllerInput.__headers_name_map__
                headers_dict: Dict[str, Header] = {
                    name: annotations_headers[name](h_value.decode())  # type: ignore
                    for h_name, h_value in headers
                    if (name := headers_map.get(h_name.decode()))  # noqa
                    in annotations_headers
                }
                kwargs.update(headers_dict)

                if middlewares:
                    middleware_request.headers = headers_dict.copy()

            if annotations_info.has_body:
                if isinstance(body_type, type) and issubclass(
                    body_type, GZipFactory
                ):
                    kwargs['body'] = body_type(value=body)

                elif isinstance(body_type, type) and issubclass(
                    body_type, str
                ):
                    kwargs['body'] = body.decode()

                else:
                    kwargs['body'] = make_json_request_body(body, body_type)

            if middlewares:
                middleware_request.body = (
                    deserialize_field('body', body_type, kwargs['body'])
                    if 'body' in kwargs
                    else None
                )

                for middleware in middlewares.pre_execution:
                    middleware(middleware_request)

                return make_asgi_input_from_requst(
                    middleware_request, ControllerInput
                )

            return as_typed_dict(  # type: ignore
                kwargs, ControllerInput
            )

        return {}

    async def build_asgi_output(
        controller_output: Any,
        status: HTTPStatus = HTTPStatus.OK,
        headers: Optional[Sequence[Header]] = None,
        content_type: Optional[ContentType] = ContentType.APPLICATION_JSON,
        return_type_: Any = None,
        middlewares: Optional[Middlewares] = None,
    ) -> ASGICallableResults:
        while iscoroutine(controller_output):
            controller_output = await controller_output

        if return_type_ is None and return_type:
            return_type_ = return_type

        if middlewares and not isinstance(controller_output, Response):
            controller_output = Response(
                body=controller_output,
                status=status,
                headers=headers,
                content_type=content_type,
            )

        if isinstance(controller_output, Response):
            if middlewares:
                for middleware in middlewares.post_execution:
                    middleware(controller_output)

            return build_asgi_output(
                controller_output['body'],
                controller_output['status'],
                controller_output['headers'],
                controller_output['content_type'],
                controller_output.__annotations__.get('body'),
            )

        elif isinstance(controller_output, dict):
            if return_type_:
                body = typed_dict_asjson(controller_output, return_type_)
            else:
                body = orjson.dumps(controller_output)

        elif (
            is_dataclass(controller_output)
            or isinstance(controller_output, tuple)
            or isinstance(controller_output, list)
        ):
            body = dataclass_asjson(controller_output)

        elif (
            isinstance(controller_output, str)
            or isinstance(controller_output, int)
            or isinstance(controller_output, bool)
            or isinstance(controller_output, float)
        ):
            if content_type == ContentType.APPLICATION_JSON:
                body = orjson.dumps(controller_output)
            else:
                body = str(controller_output).encode()

        elif controller_output is None:
            content_length = 0 if has_content_length else None

            if content_type is None and status in RESPONSES_MAP:
                if headers:
                    return (RESPONSES_MAP[status](make_asgi_headers(headers)),)

                return (RESPONSES_MAP[status](headers),)

            if headers:
                return RESPONSES_MAP[content_type](  # type: ignore
                    content_length, headers=make_asgi_headers(headers)
                )

            return RESPONSES_MAP[content_type](content_length)  # type: ignore

        else:
            raise InvalidReturnError(controller_output, controller)

        content_length = len(body) if has_content_length else None

        return (
            RESPONSES_MAP[content_type](  # type: ignore
                content_length, status, make_asgi_headers(headers)
            ),
            body,
        )

    class WrappedController(Controller):
        @functools.wraps(controller)
        async def __call__(
            self,
            path_args: ASGIPathArgs,
            query_dict: ASGIQueryDict,
            headers: ASGIHeaders,
            body: ASGIBody,
        ) -> Union[Awaitable[ASGICallableResults], ASGICallableResults]:
            try:
                print('AAAHH')
                kwargs = parse_asgi_input(
                    path_args,
                    query_dict,
                    headers,
                    body,
                    middlewares=self.middlewares,
                )
                controller_output = controller(**kwargs)
                return await build_asgi_output(
                    controller_output, middlewares=self.middlewares,
                )

            except BadRequestError as error:
                return send_bad_request_response(
                    error.dict, has_content_length, error.headers
                )

            except DeserializationError as error:
                error_dict = {
                    'name': 'field-parsing',
                    'message': error.args[0],
                }
                return send_bad_request_response(
                    error_dict, has_content_length
                )

    wrapped_controller = WrappedController()
    route = Route(
        path_pattern,
        method,
        wrapped_controller,
        annotations_info.has_path_args,
        annotations_info.has_query_dict,
        annotations_info.has_headers,
        annotations_info.has_body,
        has_options=options,
    )
    routes = [route]

    if isinstance(controller, Controller) and hasattr(controller, 'routes'):
        routes = controller.routes + routes

    wrapped_controller.routes = routes
    wrapped_controller.middlewares = route_middlewares

    return route


def make_json_request_body(body: bytes, body_type: Optional[Type[Any]]) -> Any:
    try:
        return orjson.loads(body)
    except JSONDecodeError:
        schema = (
            getattr(body_type, '__annotations__', {}) if body_type else None
        )
        schema = {k: t.__name__ for k, t in schema.items()}
        raise BadRequestError(name='invalid-body', info={'schema': schema})


def send_bad_request_response(
    error_dict: Dict[str, Any],
    has_content_length: bool,
    headers: Optional[Sequence[Header]] = None,
    content_type: ContentType = ContentType.APPLICATION_JSON,
) -> ASGICallableResults:
    body = orjson.dumps({'error': error_dict})
    content_length = len(body) if has_content_length else None

    return (
        RESPONSES_MAP[content_type](
            content_length, HTTPStatus.BAD_REQUEST, make_asgi_headers(headers)
        ),
        body,
    )


def make_asgi_headers(
    headers: Optional[Sequence[Header]],
) -> Optional[ASGIHeaders]:
    if not headers:
        return None

    return [
        (
            header.http_name.encode(),  # type: ignore
            str(header.value).encode()
            if not isinstance(header.value, bytes)
            else header.value,
        )
        for header in headers
    ]


def parse_query_args(
    name: str, type_: Type[Any], value: Optional[List[str]]
) -> Any:
    if value is not None:
        if isinstance(type_, _GenericAlias) and type_._name in (
            'List',
            'Tuple',
            'Set',
            'Deque',
        ):
            value = list(itertools.chain(*[v.split(',') for v in value]))

        elif len(value) > 1:
            raise BadRequestError(
                name='invalid-query',
                info={'name': name, 'type': type_, 'value': value},
            )
        else:
            value = value[0]  # type: ignore

        return value



import dataclasses
import re
from abc import ABC, abstractmethod
from functools import lru_cache
from typing import (
    Any,
    Callable,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Optional,
    Pattern,
)

from apidaora.exceptions import MethodNotFoundError, PathNotFoundError
from apidaora.method import MethodType

from ..middlewares import Middlewares
from .base import (
    ASGIBody,
    ASGICallableResults,
    ASGIHeaders,
    ASGIPathArgs,
    ASGIQueryDict,
)


class Controller(ABC):
    routes: List['Route']
    middlewares: Optional[Middlewares] = None

    @abstractmethod
    def __call__(
        self,
        path_args: ASGIPathArgs,
        query_dict: ASGIQueryDict,
        headers: ASGIHeaders,
        body: ASGIBody,
    ) -> ASGICallableResults:
        ...


ControllerAnnotation = Callable[
    [ASGIPathArgs, ASGIQueryDict, ASGIHeaders, ASGIBody], ASGICallableResults
]


@dataclasses.dataclass
class Route:
    path_pattern: str
    method: MethodType
    controller: Controller
    has_path_args: bool = False
    has_query: bool = False
    has_headers: bool = False
    has_body: bool = False
    has_options: bool = False


@dataclasses.dataclass
class ResolvedRoute:
    route: Route
    path_args: Dict[str, Any]
    path: str


@dataclasses.dataclass
class RoutesTreeRegex:
    name: str
    compiled_re: Optional[Pattern[Any]]


class RoutesTree(DefaultDict[str, Any]):
    regex: Optional[RoutesTreeRegex] = None

    def __init__(self) -> None:
        super().__init__(RoutesTree)


PATH_RE = re.compile(r'\{(?P<name>[^/:]+)(:(?P<pattern>[^/:]+))?\}')


def make_router(
    routes: Iterable[Route], middlewares: Optional[Middlewares] = None,
) -> Callable[[str, str], ResolvedRoute]:
    routes_tree = RoutesTree()

    for route in routes:
        if middlewares:
            if (
                not hasattr(route.controller, 'middlewares')
                or not route.controller.middlewares
            ):
                route.controller.middlewares = Middlewares(
                    **dataclasses.asdict(middlewares)
                )
            else:
                route_middlewares = Middlewares(
                    **dataclasses.asdict(route.controller.middlewares)
                )
                route_middlewares.post_routing.extend(middlewares.post_routing)
                route_middlewares.pre_execution.extend(
                    middlewares.pre_execution
                )
                route_middlewares.post_execution.extend(
                    middlewares.post_execution
                )
                route.controller.middlewares = route_middlewares

        path_pattern_parts = split_path(route.path_pattern)
        routes_tree_tmp = routes_tree

        for path_pattern_part in path_pattern_parts:
            match = PATH_RE.match(path_pattern_part)

            if match:
                group = match.groupdict()
                pattern = group.get('pattern')
                regex: Optional[Pattern[Any]] = re.compile(
                    pattern
                ) if pattern else None

                routes_tree_tmp.regex = RoutesTreeRegex(
                    group['name'], compiled_re=regex
                )
                routes_tree_tmp = routes_tree_tmp[routes_tree_tmp.regex.name]

                continue

            routes_tree_tmp = routes_tree_tmp[path_pattern_part]

        routes_tree_tmp[route.method.value] = route

    @lru_cache(maxsize=1024 * 1024)
    def route_(path: str, method: str) -> ResolvedRoute:
        path_parts = split_path(path)
        path_args = {}
        routes_tree_ = routes_tree

        for path_part in path_parts:
            if path_part in routes_tree_:
                routes_tree_ = routes_tree_[path_part]
                continue

            if routes_tree_.regex:
                compiled_re = routes_tree_.regex.compiled_re
                match = compiled_re.match(path_part) if compiled_re else None

                if not match and compiled_re:
                    raise PathNotFoundError(path)

                path_args[routes_tree_.regex.name] = path_part
                routes_tree_ = routes_tree_[routes_tree_.regex.name]

                continue

            raise PathNotFoundError(path)

        if method not in routes_tree_:
            raise MethodNotFoundError(method, path)

        return ResolvedRoute(
            route=routes_tree_[method], path_args=path_args, path=path
        )

    return route_


def split_path(path: str) -> Iterable[str]:
    return path.strip(' /').split('/')



from typing import Any, Dict

from jsondaora import jsondaora

from apidaora import (
    CorsMiddleware,
    Header,
    MiddlewareRequest,
    Middlewares,
    Response,
    appdaora,
    route,
    text,
)


def post_routing_middleware(
    path_pattern: str, path_args: Dict[str, Any]
) -> None:
    path_args['name'] = path_args['name'].replace('Me', 'You')


def pre_execution_middleware(request: MiddlewareRequest) -> None:
    if request.body:
        request.body.name = request.body.name.replace('Me', 'You')


def post_execution_middleware(response: Response) -> None:
    response.headers = [
        PostExecutionHeader(
            len(response.body.replace('Hello ', '').replace('!', ''))
        )
    ]


# The route middlewares takes precedence over general middlewares


@route.get(
    '/hello-post-routing/{name}',
    middlewares=Middlewares(post_routing=[post_routing_middleware]),
)
async def post_routing_middleware_controller(name: str) -> str:
    return hello(name)


@jsondaora
class PreExecutionBody:
    name: str


@route.post(
    '/hello-pre-execution/',
    middlewares=Middlewares(pre_execution=[pre_execution_middleware]),
)
async def pre_execution_middleware_controller(body: PreExecutionBody) -> str:
    return hello(body.name)


class PostExecutionHeader(Header, type=int, http_name='x-name-len'):
    ...


@route.get(
    '/hello-post-execution',
    middlewares=Middlewares(post_execution=[post_execution_middleware]),
)
async def post_execution_middleware_controller(name: str) -> Response:
    return text(body=hello(name))


def hello(name: str) -> str:
    return f'Hello {name}!'


app = appdaora(
    [
        post_routing_middleware_controller,
        pre_execution_middleware_controller,
        post_execution_middleware_controller,
    ],
    middlewares=Middlewares(
        post_execution=[CorsMiddleware(servers_all='my-server.domain')]
    ),
)



from typing import TypedDict

from jsondaora import IntegerField, StringField, jsondaora

from apidaora import Header, appdaora, route


class Integer(IntegerField, minimum=18):
    ...


class String(StringField, max_length=100):
    ...


class Age(Header, type=Integer):
    ...


@jsondaora
class You(TypedDict):
    name: str
    last_name: str
    location: str
    age: int


@jsondaora
class ReqBody(TypedDict):
    last_name: str


@jsondaora
class HelloOutput(TypedDict):
    hello_message: str
    about_you: You


@route.put('/hello/{name}')
async def hello_controller(
    name: str, location: String, age: Age, body: ReqBody
) -> HelloOutput:
    you = You(
        name=name,
        location=location.value,
        age=age.value.value,
        last_name=body['last_name'],
    )
    return HelloOutput(
        hello_message=await hello_message(name, location.value), about_you=you
    )


async def hello_message(name: str, location: str) -> str:
    return f'Hello {name}! Welcome to {location}!'


app = appdaora(hello_controller)



import time

from apidaora import appdaora, route


@route.background('/hello-single', single_running=True)
def hello_task(name: str) -> str:
    time.sleep(1)
    return f'Hello {name}!'


app = appdaora(hello_task)



from http import HTTPStatus
from typing import Awaitable, Optional, Tuple

from ..content import ContentType
from .base import ASGIHeaders, ASGIResponse, Sender


HTTP_RESPONSE_START = 'http.response.start'
JSON_CONTENT_HEADER = (
    b'content-type',
    ContentType.APPLICATION_JSON.value.encode(),
)
HTML_CONTENT_HEADER = (b'content-type', ContentType.TEXT_HTML.value.encode())
PLAINTEXT_CONTENT_HEADER = (
    b'content-type',
    ContentType.TEXT_PLAIN.value.encode(),
)

JSON_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.OK.value,
    'headers': [JSON_CONTENT_HEADER],
}

HTML_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.OK.value,
    'headers': [HTML_CONTENT_HEADER],
}

PLAINTEXT_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.OK.value,
    'headers': [PLAINTEXT_CONTENT_HEADER],
}

NOT_FOUND_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.NOT_FOUND.value,
    'headers': [],
}

METHOD_NOT_ALLOWED_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.METHOD_NOT_ALLOWED.value,
    'headers': [],
}

NO_CONTENT_RESPONSE: ASGIResponse = {
    'type': HTTP_RESPONSE_START,
    'status': HTTPStatus.NO_CONTENT.value,
    'headers': [],
}


def make_response(
    content_length: Optional[int],
    status: HTTPStatus,
    headers: Optional[ASGIHeaders],
    default_value: ASGIResponse,
    default_content_header: Optional[Tuple[bytes, bytes]],
) -> ASGIResponse:
    default_headers: ASGIHeaders

    if content_length is None and headers is None:
        return default_value

    if content_length and default_content_header:
        default_headers = [
            default_content_header,
            (b'content-length', str(content_length).encode()),
        ]
    elif default_content_header:
        default_headers = [default_content_header]
    else:
        default_headers = []

    if headers and default_headers:
        if isinstance(headers, list):
            headers = default_headers + headers
        else:
            headers = default_headers + list(headers)

    return {
        'type': HTTP_RESPONSE_START,
        'status': status.value,
        'headers': default_headers if not headers else headers,
    }


def make_json_response(
    content_length: Optional[int] = None,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        content_length, status, headers, JSON_RESPONSE, JSON_CONTENT_HEADER
    )


def make_text_response(
    content_length: Optional[int] = None,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        content_length,
        status,
        headers,
        PLAINTEXT_RESPONSE,
        PLAINTEXT_CONTENT_HEADER,
    )


def make_html_response(
    content_length: Optional[int] = None,
    status: HTTPStatus = HTTPStatus.OK,
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        content_length, status, headers, HTML_RESPONSE, HTML_CONTENT_HEADER
    )


def make_no_content_response(
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        None, HTTPStatus.NO_CONTENT, headers, NO_CONTENT_RESPONSE, None,
    )


def make_not_found_response(
    headers: Optional[ASGIHeaders] = None,
) -> ASGIResponse:
    return make_response(
        None, HTTPStatus.NOT_FOUND, headers, NOT_FOUND_RESPONSE, None,
    )


async def send_response(
    send: Sender, response: ASGIResponse, body: bytes
) -> None:
    await send(response)  # type: ignore
    await send(
        {'type': 'http.response.body', 'body': body, 'more_body': False}
    )


def send_not_found(send: Sender) -> Awaitable[None]:
    return send_response(send, NOT_FOUND_RESPONSE, b'')


def send_method_not_allowed_response(send: Sender) -> Awaitable[None]:
    return send_response(send, METHOD_NOT_ALLOWED_RESPONSE, b'')



from apidaora import GZipFactory, appdaora, route


class GzipBody(GZipFactory):
    mode = 'rt'


@route.post('/hello')
def gzip_hello(body: GzipBody) -> str:
    with body.open() as file:
        return file.read()


app = appdaora(gzip_hello)



from typing import Callable, List

from .header import Header
from .method import MethodType
from .responses import Response, no_content


class AllowHeader(Header, type=str, http_name='allow'):
    ...


def make_options_controller(
    methods: List[MethodType],
) -> Callable[[], Response]:
    def controller() -> Response:
        return no_content(
            headers=[AllowHeader(','.join([m.value for m in methods]))]
        )

    return controller



from typing import List

from apidaora import Middlewares, PreventRequestMiddleware, appdaora, route


prevent_request_middleware = PreventRequestMiddleware()


@route.post(
    '/prevent-request',
    middlewares=Middlewares(
        post_routing=[prevent_request_middleware.set_in_process],
        pre_execution=[prevent_request_middleware.remove_in_process],
    ),
)
async def pre_execution_middleware_controller(body: List[int]) -> int:
    return len(body)


app = appdaora(pre_execution_middleware_controller)



from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence

from .header import Header


class APIDaoraError(Exception):
    ...


class MethodNotFoundError(APIDaoraError):
    ...


class PathNotFoundError(APIDaoraError):
    ...


class InvalidReturnError(APIDaoraError):
    def __str__(self) -> str:
        return (
            f"handler_name='{self.args[1].__name__}', "
            f"return_type='{type(self.args[0]).__name__}', "
            f"return_value='{self.args[0]}'"
        )


@dataclass
class BadRequestError(APIDaoraError):
    name: str
    info: Dict[str, Any]
    headers: Optional[Sequence[Header]] = None

    def __str__(self) -> str:
        return f"name='{self.name}', info={self.info}"

    @property
    def dict(self) -> Dict[str, Any]:
        return {'name': self.name, 'info': self.info}


class InvalidTasksRepositoryError(APIDaoraError):
    ...


class InvalidRouteArgumentsError(APIDaoraError):
    ...



from ..app import appdaora
from ..asgi.responses import (
    HTML_RESPONSE,
    JSON_RESPONSE,
    METHOD_NOT_ALLOWED_RESPONSE,
    NOT_FOUND_RESPONSE,
    NO_CONTENT_RESPONSE,
    PLAINTEXT_RESPONSE,
)
from .decorator import route


appdaora_core = appdaora


__all__ = [
    'appdaora_core',
    'route',
    'JSON_RESPONSE',
    'HTML_RESPONSE',
    'PLAINTEXT_RESPONSE',
    'NOT_FOUND_RESPONSE',
    'METHOD_NOT_ALLOWED_RESPONSE',
    'NO_CONTENT_RESPONSE',
]
